from pydantic import BaseModel, Field
from typing import Annotated, Union, List, Any, Literal
from datapipeline.config.dataset.normalize import floor_time_to_resolution


class GroupKey(BaseModel):
    type: str
    field: str

    def normalize(self, val: Any) -> Any:
        return val


class TimeKey(GroupKey):
    type: Literal["time"]
    resolution: str = Field(..., pattern=r"^\d+(min|h)$")

    def normalize(self, val: Any) -> Any:
        return floor_time_to_resolution(val, self.resolution)


class CategoricalKey(GroupKey):
    type: Literal["category"]


GroupKeyUnion = Annotated[Union[TimeKey,
                                CategoricalKey], Field(discriminator="type")]


class GroupBy(BaseModel):
    keys: List[GroupKeyUnion]




from typing import Literal
from datapipeline.config.dataset.dataset import RecordDatasetConfig, FeatureDatasetConfig
from datapipeline.services.bootstrap import _load_by_key  # your helper

Stage = Literal["records", "features", "vectors"]


def load_dataset(project_yaml, stage: Stage):
    ds_doc = _load_by_key(project_yaml, "dataset")

    if stage == "records":
        return RecordDatasetConfig.model_validate(ds_doc)
    elif stage == "features":
        return FeatureDatasetConfig.model_validate(ds_doc)
    elif stage == "vectors":
        return FeatureDatasetConfig.model_validate(ds_doc)
    else:
        raise ValueError(f"Unknown stage: {stage}")



from typing import Iterator, Generic, TypeVar, Optional
from .base import SourceInterface
from .loader import RawDataLoader
from .parser import DataParser

TRecord = TypeVar("TRecord")


class Source(SourceInterface[TRecord], Generic[TRecord]):
    def __init__(self, loader: RawDataLoader, parser: DataParser[TRecord]):
        self.loader = loader
        self.parser = parser

    def stream(self) -> Iterator[TRecord]:
        for row in self.loader.load():
            try:
                parsed = self.parser.parse(row)
                if parsed is not None:
                    yield parsed
            except Exception:
                continue

    def count(self) -> Optional[int]:
        try:
            return self.loader.count()
        except Exception:
            return None





"""Composable pipeline stages used by the runtime."""

from __future__ import annotations

from collections import defaultdict
from itertools import groupby
from typing import Any, Iterable, Iterator, Optional, Sequence, Tuple

from datapipeline.config.dataset.feature import FeatureRecordConfig
from datapipeline.config.dataset.group_by import GroupBy
from datapipeline.domain.feature import FeatureRecord
from datapipeline.domain.vector import Vector, vectorize_record_group
from datapipeline.pipeline.utils.memory_sort import memory_sorted
from datapipeline.pipeline.utils.ordering import canonical_key
from datapipeline.pipeline.utils.transform_utils import (
    filter_record_stream,
    record_to_feature,
    transform_feature_stream,
    transform_record_stream,
)


def record_stage(
    raw_stream: Iterable[Any],
    filters: Optional[Sequence[dict[str, Any]]] = None,
    transforms: Optional[Sequence[dict[str, Any]]] = None,
) -> Iterator[Any]:
    """Apply configured filters and transforms to the raw record stream."""

    stream = filter_record_stream(iter(raw_stream), filters)
    return transform_record_stream(stream, transforms)


def feature_stage(
    record_stream: Iterable[Any],
    cfg: FeatureRecordConfig,
    group_by: GroupBy,
) -> Iterator[FeatureRecord]:
    """Wrap filtered records as FeatureRecord objects.
    Assign partition-aware feature_ids and normalized group_keys before transforms.
    Sort feature streams, apply feature/sequence transforms, and emit canonical order."""

    stream = record_to_feature(record_stream, cfg, group_by)
    stream = memory_sorted(
        stream,
        batch_size=100000,
        key=lambda fr: (fr.feature_id, getattr(fr.record, "time", 0)),
    )
    stream = transform_feature_stream(stream, cfg)
    return memory_sorted(stream, batch_size=100000, key=canonical_key)


def vector_stage(merged: Iterator[FeatureRecord]) -> Iterator[Tuple[Any, Vector]]:
    """Group the merged feature stream by group_key.
    Coalesce each partitioned feature_id into record buckets.
    Yield (group_key, Vector) pairs ready for downstream consumption."""

    for group_key, group in groupby(merged, key=lambda fr: fr.group_key):
        feature_map = defaultdict(list)
        for fr in group:
            records = fr.record if isinstance(fr.record, list) else [fr.record]
            feature_map[fr.feature_id].extend(records)
        yield group_key, vectorize_record_group(feature_map)



import importlib.metadata as md
from functools import lru_cache
import yaml
from pathlib import Path


@lru_cache
def load_ep(group: str, name: str):
    eps = md.entry_points().select(group=group, name=name)
    if not eps:
        available = ", ".join(
            sorted(ep.name for ep in md.entry_points().select(group=group)))
        raise ValueError(
            f"No entry point '{name}' in '{group}'. Available: {available or '(none)'}")
    if len(eps) > 1:
        mods = ", ".join(f"{ep.module}:{ep.attr}" for ep in eps)
        raise ValueError(
            f"Ambiguous entry point '{name}' in '{group}': {mods}")
    return eps[0].load()


def load_yaml(p: Path) -> dict:
    try:
        with p.open("r", encoding="utf-8") as f:
            data = yaml.safe_load(f)
    except FileNotFoundError as e:
        raise FileNotFoundError(f"YAML file not found: {p}") from e
    except yaml.YAMLError as e:
        raise ValueError(f"Invalid YAML in {p}: {e}") from e

    if data is None:
        return {}
    if not isinstance(data, dict):
        raise TypeError(
            f"Top-level YAML in {p} must be a mapping, got {type(data).__name__}")
    return data



from pathlib import Path
from datapipeline.services.scaffold.plugin import scaffold_plugin


def station(subcmd: str, name: str | None, out: str) -> None:
    if subcmd == "init":
        if not name:
            print("❗ --name is required for station init")
            raise SystemExit(2)
        scaffold_plugin(name, Path(out))



from datapipeline.domain.record import TimeFeatureRecord
from datetime import timedelta
from typing import Iterator
from datapipeline.utils.time import parse_timecode


def shift_record_time(record: TimeFeatureRecord, lag: timedelta) -> TimeFeatureRecord:
    record.time = record.time - lag
    return record


def time_lag(stream: Iterator[TimeFeatureRecord], lag: str) -> Iterator[TimeFeatureRecord]:
    lag_td = parse_timecode(lag)
    for record in stream:
        yield shift_record_time(record, lag_td)



from datapipeline.services.scaffold.domain import create_domain


def handle(subcmd: str, domain: str | None, time_aware: bool) -> None:
    if subcmd in {"create", "add"}:
        if not domain:
            print("❗ --domain is required")
            raise SystemExit(2)
        create_domain(domain=domain, time_aware=time_aware, root=None)



from pathlib import Path
import pkg_resources

from ..constants import COMPOSED_LOADER_EP

SKELETON_DIR = Path(pkg_resources.resource_filename(
    "datapipeline", "templates/plugin_skeleton"))


def scaffold_plugin(name: str, outdir: Path) -> None:
    target = (outdir / name).absolute()
    if target.exists():
        print(f"❗ `{target}` already exists")
        raise SystemExit(1)
    import shutil
    shutil.copytree(SKELETON_DIR, target)
    pkg_dir = target / "src" / "{{PACKAGE_NAME}}"
    pkg_dir.rename(target / "src" / name)
    for p in (target / "pyproject.toml", target / "README.md"):
        text = p.read_text().replace("{{PACKAGE_NAME}}", name)
        text = text.replace("{{COMPOSED_LOADER_EP}}", COMPOSED_LOADER_EP)
        p.write_text(text)
    print(f"✨ Created plugin skeleton at {target}")



from __future__ import annotations

import json
import pickle
import sys
from itertools import islice
from pathlib import Path
from typing import Iterator, Optional, Tuple

from tqdm import tqdm

from datapipeline.cli.openers import open_canonical_stream_visual
from datapipeline.config.dataset.dataset import FeatureDatasetConfig, RecordDatasetConfig
from datapipeline.config.dataset.loader import load_dataset
from datapipeline.pipeline.pipelines import (
    build_feature_pipeline,
    build_record_pipeline,
    build_vector_pipeline,
)
from datapipeline.services.bootstrap import bootstrap
from datapipeline.streams.canonical import open_canonical_stream
from datapipeline.domain.vector import Vector


def _print_head(iterable: Iterator[object], limit: int) -> int:
    count = 0
    try:
        for item in iterable:
            tqdm.write(str(item))
            count += 1
            if count >= limit:
                break
    except KeyboardInterrupt:
        pass
    return count


def _run_records(dataset: RecordDatasetConfig, limit: int) -> None:
    for cfg in dataset.features:
        print(f"\n🍷 pouring records for {cfg.feature_id}")
        records = build_record_pipeline(cfg, open_canonical_stream_visual)
        printed = _print_head(records, limit)
        print(f"(poured {printed} records)")


def _run_features(dataset: FeatureDatasetConfig, limit: int) -> None:
    group_by = dataset.group_by
    for cfg in dataset.features:
        feature_id = getattr(cfg, "feature_id", "?")
        print(f"\n🛠️ building features for {feature_id}")
        features = build_feature_pipeline(
            cfg, group_by, open_canonical_stream_visual)
        printed = _print_head(features, limit)
        tqdm.write(f"(built {printed} feature records)")


def _run_vectors(dataset: FeatureDatasetConfig, limit: int) -> None:
    print("\n🥄 stirring vectors")
    vectors = build_vector_pipeline(
        dataset.features, dataset.group_by, open_canonical_stream_visual)
    printed = _print_head(vectors, limit)
    print(f"(stirred {printed} vectors)")


def handle_prep(action: str, project: str, limit: int = 20) -> None:
    stage_lookup = {"pour": "records", "build": "features", "stir": "vectors"}
    if action not in stage_lookup:
        raise ValueError(f"Unknown prep action: {action}")

    project_path = Path(project)
    dataset = load_dataset(project_path, stage_lookup[action])
    bootstrap(project_path)

    features = list(dataset.features or [])
    if not features:
        print("(no features configured; nothing to prep)")
        return

    if action == "pour":
        _run_records(dataset, limit)
    elif action == "build":
        _run_features(dataset, limit)
    else:
        _run_vectors(dataset, limit)


def _limit_vectors(vectors: Iterator[Tuple[object, Vector]], limit: Optional[int]) -> Iterator[Tuple[object, Vector]]:
    if limit is None:
        yield from vectors
    else:
        yield from islice(vectors, limit)


def _serve_print(vectors: Iterator[Tuple[object, Vector]], limit: Optional[int]) -> None:
    count = 0
    try:
        for group_key, vector in _limit_vectors(vectors, limit):
            print(f"group={group_key}: {vector.values}")
            count += 1
    except KeyboardInterrupt:
        pass
    print(f"(served {count} vectors to stdout)")


def _serve_stream(vectors: Iterator[Tuple[object, Vector]], limit: Optional[int]) -> None:
    count = 0
    try:
        for group_key, vector in _limit_vectors(vectors, limit):
            payload = {"group": list(group_key) if isinstance(group_key, tuple) else group_key,
                       "values": vector.values}
            print(json.dumps(payload, default=str))
            count += 1
    except KeyboardInterrupt:
        pass
    print(f"(streamed {count} vectors)", file=sys.stderr)


def _serve_pt(vectors: Iterator[Tuple[object, Vector]], limit: Optional[int], destination: Path) -> None:
    data = []
    for group_key, vector in _limit_vectors(vectors, limit):
        normalized_key = list(group_key) if isinstance(
            group_key, tuple) else group_key
        data.append((normalized_key, vector.values))
    destination.parent.mkdir(parents=True, exist_ok=True)
    with destination.open("wb") as fh:
        pickle.dump(data, fh)
    print(f"💾 Saved {len(data)} vectors to {destination}")


def handle_serve(project: str, limit: Optional[int], output: str) -> None:
    project_path = Path(project)
    dataset = load_dataset(project_path, "vectors")
    bootstrap(project_path)

    features = list(dataset.features or [])
    if not features:
        print("(no features configured; nothing to serve)")
        return

    vectors = build_vector_pipeline(
        dataset.features, dataset.group_by, open_canonical_stream)

    if output == "print":
        _serve_print(vectors, limit)
    elif output == "stream":
        _serve_stream(vectors, limit)
    elif output.endswith(".pt"):
        _serve_pt(vectors, limit, Path(output))
    else:
        print("❗ Unsupported output format. Use 'print', 'stream', or a .pt file path.")
        raise SystemExit(2)



import os

PARSERS_EP = os.getenv("DP_PARSERS_EP",    "datapipeline.parsers")
LOADERS_EP = os.getenv("DP_LOADERS_EP",    "datapipeline.loaders")
MAPPERS_EP = os.getenv("DP_MAPPERS_EP",    "datapipeline.mappers")
FILTERS_EP = os.getenv("DP_FILTERS_EP",    "datapipeline.filters")
TRANSFORMS_EP = os.getenv("DP_TRANSFORMS_EP", "datapipeline.transforms")



PARSER_KEY = "parser"
LOADER_KEY = "loader"
SOURCE_KEY = "source"
MAPPER_KEY = "mapper"
ENTRYPOINT_KEY = "entrypoint"
ARGS_KEY = "args"

PARSERS_GROUP = "parsers"
LOADERS_GROUP = "loaders"
MAPPERS_GROUP = "mappers"
FILTERS_GROUP = "filters"
COMPOSED_LOADER_EP = "composed.loader"



from collections import defaultdict, Counter
from typing import Any, Hashable, Iterable

import numpy as np


class VectorStatsCollector:
    def __init__(self, expected_feature_ids: Iterable[str]):
        self.expected_feature_ids = set(expected_feature_ids)
        self.missing_features = Counter()
        self.empty_vectors = 0
        self.total_vectors = 0
        self.per_group_missing = defaultdict(set)

    def update(self, group_key: Hashable, feature_vector: dict[str, Any]):
        self.total_vectors += 1

        present_features = set(feature_vector.keys())

        if not present_features:
            self.empty_vectors += 1

        missing = self.expected_feature_ids - present_features
        for feature_id in missing:
            self.missing_features[feature_id] += 1
            self.per_group_missing[group_key].add(feature_id)

        # 🔍 Check for features present but with missing/invalid values
        for fid in present_features & self.expected_feature_ids:
            val = feature_vector[fid]
            if val is None or (isinstance(val, float) and np.isnan(val)):
                self.missing_features[fid] += 1
                self.per_group_missing[group_key].add(fid)

    def print_report(self):
        print("\n=== Vector Quality Report ===")
        print(f"Total vectors processed: {self.total_vectors}")
        print(f"Empty vectors: {self.empty_vectors}")
        print(
            f"Features expected: {sorted(self.expected_feature_ids)[:10] }... (total {len(self.expected_feature_ids)})")

        print("\n→ Missing feature counts:")
        for fid, count in sorted(self.missing_features.items(), key=lambda x: -x[1]):
            print(f"  - {fid}: missing in {count} vectors")

        if self.per_group_missing:
            print("\n→ Groups with missing features (sample):")
            for group_key, missing in list(self.per_group_missing.items())[:5]:
                print(f"  - Group {group_key}: {sorted(missing)}")



from __future__ import annotations

from pathlib import Path
from typing import Mapping
import re
from datetime import datetime, timezone

from datapipeline.utils.load import load_yaml
from datapipeline.config.catalog import StreamsConfig
from datapipeline.config.project import ProjectConfig
from datapipeline.services.project_paths import streams_dir, sources_dir
from datapipeline.services.constants import (
    PARSER_KEY,
    LOADER_KEY,
    SOURCE_KEY,
    MAPPER_KEY,
    ENTRYPOINT_KEY,
)
from datapipeline.streams.raw import register_source
from datapipeline.streams.canonical import register_stream
from datapipeline.services.factories import (
    build_source_from_spec,
    build_mapper_from_spec,
)

SRC_PARSER_KEY = PARSER_KEY
SRC_LOADER_KEY = LOADER_KEY


def _project(project_yaml: Path) -> ProjectConfig:
    """Load and validate project.yaml."""
    data = load_yaml(project_yaml)
    return ProjectConfig.model_validate(data)


def _paths(project_yaml: Path) -> Mapping[str, str]:
    proj = _project(project_yaml)
    return proj.paths.model_dump()


def _load_by_key(project_yaml: Path, key: str) -> dict:
    """Load a YAML document referenced by project.paths[key]. (Legacy)"""
    p = _paths(project_yaml).get(key)
    if not p:
        raise FileNotFoundError(f"project.paths must include '{key}'.")
    return load_yaml(Path(p))


def _globals(project_yaml: Path) -> dict[str, str]:
    """Return project-level globals for interpolation.

    If a value is a datetime, normalize to strict UTC Z-format string so
    downstream components expecting ISO Z will work predictably.
    Otherwise, coerce to string.
    """
    proj = _project(project_yaml)
    g = proj.globals.model_dump()
    out: dict[str, str] = {}
    for k, v in g.items():
        if isinstance(v, datetime):
            v = v.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        out[str(k)] = str(v)
    return out


_VAR_RE = re.compile(r"\$\{([^}]+)\}")


def _interpolate(obj, vars_: dict[str, str]):
    """Recursively substitute ${var} in strings using vars_ map.

    Minimal behavior: if a key is missing, leave placeholder as-is.
    """
    if isinstance(obj, dict):
        return {k: _interpolate(v, vars_) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_interpolate(v, vars_) for v in obj]
    if isinstance(obj, str):
        def repl(m):
            key = m.group(1)
            return vars_.get(key, m.group(0))
        return _VAR_RE.sub(repl, obj)
    return obj


def _load_sources_from_dir(project_yaml: Path, vars_: dict[str, str]) -> dict:
    """Aggregate per-source YAML files into a raw-sources mapping.

    Expects each file to define a single source with top-level 'parser' and
    'loader' keys. The source alias is inferred from the filename (without
    extension).
    """
    import os
    src_dir = sources_dir(project_yaml)
    if not src_dir.exists() or not src_dir.is_dir():
        return {}
    out: dict[str, dict] = {}
    for fname in sorted(os.listdir(src_dir)):
        if not (fname.endswith(".yaml") or fname.endswith(".yml")):
            continue
        data = load_yaml(src_dir / fname)
        if not isinstance(data, dict):
            continue
        # Top-level parser/loader; alias inferred from filename
        if isinstance(data.get(SRC_PARSER_KEY), dict) and isinstance(data.get(SRC_LOADER_KEY), dict):
            alias = Path(fname).stem
            out[alias] = _interpolate(data, vars_)
            continue
    return out


def _load_canonical_streams(project_yaml: Path, vars_: dict[str, str]) -> dict:
    """Aggregate canonical stream specs from streams_dir (supports subfolders).

    Recursively scans for *.yml|*.yaml under the configured streams dir.
    Stream alias is derived from the relative path with '/' replaced by '.'
    and extension removed, e.g. 'metobs/precip.yaml' → 'metobs.precip'.
    """
    out: dict[str, dict] = {}
    sdir = streams_dir(project_yaml)
    if not sdir.exists() or not sdir.is_dir():
        return {}
    for p in sorted(sdir.rglob("*.y*ml")):
        if not p.is_file():
            continue
        data = load_yaml(p)
        if isinstance(data, dict) and (SOURCE_KEY in data):
            m = data.get(MAPPER_KEY)
            if (not isinstance(m, dict)) or (ENTRYPOINT_KEY not in (m or {})):
                data[MAPPER_KEY] = None
            rel = p.relative_to(sdir)
            alias = rel.with_suffix("").as_posix().replace("/", ".")
            out[alias] = _interpolate(data, vars_)
    return out


def load_streams(project_yaml: Path) -> StreamsConfig:
    vars_ = _globals(project_yaml)
    raw = _load_sources_from_dir(project_yaml, vars_)
    canonical = _load_canonical_streams(project_yaml, vars_)
    return StreamsConfig(raw=raw, canonical=canonical)


def init_streams(cfg: StreamsConfig) -> None:
    """Compile typed streams config into runtime registries."""
    for alias, spec in (cfg.raw or {}).items():
        register_source(alias, build_source_from_spec(spec))
    for alias, spec in (cfg.canonical or {}).items():
        fn, margs = build_mapper_from_spec(spec.mapper)  # None -> identity
        register_stream(alias, source_alias=spec.source,
                        mapper=fn, mapper_args=margs)


def bootstrap(project_yaml: Path) -> StreamsConfig:
    """One-call init: load streams.yaml and register raw/canonical streams."""
    streams = load_streams(project_yaml)
    init_streams(streams)
    return streams



from typing import Iterable, Iterator, Callable, TypeVar
import heapq


def apply_pipeline(stream, stages):
    for stage in stages:
        stream = stage(stream)
    return stream


T = TypeVar("T")


def read_batches(iterable: Iterable[T], batch_size: int, key: Callable[[T], any]) -> Iterator[list[T]]:
    batch = []
    for item in iterable:
        batch.append(item)
        if len(batch) == batch_size:
            yield sorted(batch, key=key)
            batch = []
    if batch:
        yield sorted(batch, key=key)


def memory_sorted(iterable: Iterable[T], batch_size: int, key: Callable[[T], any]) -> Iterator[T]:
    sorted_batches = read_batches(iterable, batch_size, key)
    return heapq.merge(*sorted_batches, key=key)



from typing import NamedTuple, Callable, Iterator, Any, Dict, Tuple
from datapipeline.streams.raw import open_raw_stream
_CANON: Dict[str, Tuple[str, Callable[..., Iterator[Any]], dict]] = {}


def register_stream(alias: str, source_alias: str, mapper: Callable[..., Iterator[Any]], mapper_args: dict | None = None) -> None:
    _CANON[alias] = (source_alias, mapper, mapper_args or {})


class CanonicalEntry(NamedTuple):
    source_alias: str
    mapper: Callable[..., Iterator[Any]]
    mapper_args: dict


def canonical_entry(alias: str) -> CanonicalEntry:
    src_alias, mapper, args = _CANON[alias]
    return CanonicalEntry(src_alias, mapper, dict(args))


def open_canonical_stream(alias: str) -> Iterator[Any]:
    try:
        source_alias, mapper, kwargs = _CANON[alias]
    except KeyError:
        raise ValueError(
            f"Unknown stream alias '{alias}'. Available: {', '.join(sorted(_CANON))}"
        )
    return mapper(open_raw_stream(source_alias), **kwargs)



from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from typing import Any, Optional, Tuple

from datapipeline.config.dataset.feature import BaseRecordConfig
from datapipeline.config.dataset.group_by import GroupBy
from datapipeline.domain.feature import FeatureRecord
from datapipeline.domain.record import Record
from datapipeline.pipeline.utils.keygen import RecordKeyGenerator
from datapipeline.plugins import FILTERS_EP, TRANSFORMS_EP
from datapipeline.utils.load import load_ep


def _extract_single_pair(clause: Mapping[str, Any], kind: str) -> Tuple[str, Any]:
    """Validate that *clause* is a one-key mapping and return that pair."""

    if not isinstance(clause, Mapping) or len(clause) != 1:
        raise TypeError(f"{kind} must be one-key mapping, got: {clause!r}")
    return next(iter(clause.items()))


def _call_with_params(fn: Callable, stream: Iterator[Any], params: Any) -> Iterator[Any]:
    """Invoke an entry-point callable with optional params semantics."""

    if params is None:
        return fn(stream)
    if isinstance(params, (list, tuple)):
        return fn(stream, *params)
    if isinstance(params, Mapping):
        return fn(stream, **params)
    return fn(stream, params)


def _instantiate_entry_point(cls: Callable[..., Any], params: Any) -> Any:
    """Instantiate a transform class with parameters from the config."""

    if params is None:
        return cls()
    if isinstance(params, Mapping):
        return cls(**params)
    if isinstance(params, (list, tuple)):
        return cls(*params)
    return cls(params)


def filter_record_stream(
    stream: Iterator[Any],
    filters: Optional[Sequence[Mapping[str, Any]]],
) -> Iterator[Any]:
    """Apply all configured filters to the record stream."""

    for clause in filters or ():
        name, mapping = _extract_single_pair(clause, "Filter")
        field, target = _extract_single_pair(mapping, f"Filter '{name}'")
        fn: Callable = load_ep(group=FILTERS_EP, name=name)
        stream = fn(stream, field, target)
    return stream


def transform_record_stream(
    stream: Iterator[Any],
    transforms: Optional[Sequence[Mapping[str, Any]]],
) -> Iterator[Any]:
    """Apply configured record-level transforms to the stream."""

    for clause in transforms or ():
        name, params = _extract_single_pair(clause, "Transform")
        fn: Callable = load_ep(group=TRANSFORMS_EP, name=name)
        stream = _call_with_params(fn, stream, params)
    return stream


def _instantiate_transforms(
    group: str, clauses: Optional[Sequence[Mapping[str, Any]]]
) -> list[Any]:
    """Instantiate configured transform classes for later application."""

    instances: list[Any] = []
    for clause in clauses or ():
        name, params = _extract_single_pair(clause, "Transform")
        cls = load_ep(group=group, name=name)
        instances.append(_instantiate_entry_point(cls, params))
    return instances


def transform_feature_stream(
    stream: Iterator[FeatureRecord],
    config: BaseRecordConfig,
) -> Iterator[FeatureRecord]:
    """Apply feature and sequence transforms declared in the config."""

    feature_tf = getattr(config, "feature_transforms", None) or []
    for transform in _instantiate_transforms("datapipeline.transforms.feature", feature_tf):
        stream = transform.apply(stream)

    seq_tf = getattr(config, "sequence_transforms", None) or []
    for transform in _instantiate_transforms("datapipeline.transforms.sequence", seq_tf):
        stream = transform.apply(stream)
    return stream


def record_to_feature(
    stream: Iterable[Record],
    config: BaseRecordConfig,
    group_by: GroupBy,
) -> Iterator[FeatureRecord]:
    """Convert raw records into ``FeatureRecord`` instances."""

    keygen = RecordKeyGenerator(config.partition_by)

    def group_key(rec: Record) -> tuple:
        return tuple(k.normalize(getattr(rec, k.field)) for k in group_by.keys)

    for rec in stream:
        yield FeatureRecord(
            record=rec,
            feature_id=keygen.generate(config.feature_id, rec),
            group_key=group_key(rec),
        )



import sys
from datapipeline.services.paths import pkg_root, resolve_base_pkg_dir
from datapipeline.services.entrypoints import read_group_entries
import yaml
from datapipeline.services.constants import FILTERS_GROUP, MAPPER_KEY, ENTRYPOINT_KEY, ARGS_KEY, SOURCE_KEY
from datapipeline.services.project_paths import sources_dir as resolve_sources_dir, streams_dir as resolve_streams_dir
from datapipeline.services.scaffold.mappers import attach_source_to_domain
import re


def _pick_from_list(prompt: str, options: list[str]) -> str:
    print(prompt, file=sys.stderr)
    for i, opt in enumerate(options, 1):
        print(f"  [{i}] {opt}", file=sys.stderr)
    while True:
        sel = input("> ").strip()
        if sel.isdigit():
            idx = int(sel)
            if 1 <= idx <= len(options):
                return options[idx - 1]
        print("Please enter a number from the list.", file=sys.stderr)


def handle(time_aware: bool) -> None:
    root_dir, name, pyproject = pkg_root(None)

    # Discover sources by scanning sources_dir YAMLs
    proj_path = root_dir / "config" / "project.yaml"
    sources_dir = resolve_sources_dir(proj_path)
    source_options = []
    if sources_dir.exists():
        source_options = sorted(p.stem for p in sources_dir.glob("*.y*ml"))
    if not source_options:
        print("❗ No sources found. Create one first (jerry distillery add ...)")
        raise SystemExit(2)

    src_key = _pick_from_list("Select a source to link:", source_options)
    # Expect aliases from sources_dir filenames: provider_dataset.yaml
    parts = src_key.split("_", 1)
    if len(parts) != 2:
        print("❗ Source alias must be 'provider_dataset' (from sources/<alias>.yaml)", file=sys.stderr)
        raise SystemExit(2)
    provider, dataset = parts[0], parts[1]

    # Discover domains by scanning the package, fallback to EPs if needed
    base = resolve_base_pkg_dir(root_dir, name)
    domain_options = []
    for dirname in ("domains",):
        dom_dir = base / dirname
        if dom_dir.exists():
            domain_options.extend(
                [p.name for p in dom_dir.iterdir() if p.is_dir()
                 and (p / "model.py").exists()]
            )
    domain_options = sorted(set(domain_options))
    if not domain_options:
        domain_options = sorted(
            read_group_entries(pyproject, FILTERS_GROUP).keys())
    if not domain_options:
        print("❗ No domains found. Create one first (jerry spirit add ...)")
        raise SystemExit(2)

    dom_name = _pick_from_list("Select a domain to link to:", domain_options)

    # create mapper + EP (domain.origin)
    attach_source_to_domain(domain=dom_name, provider=provider,
                            dataset=dataset, time_aware=time_aware, root=None)

    def _slug(s: str) -> str:
        s = s.strip().lower()
        s = re.sub(r"[^a-z0-9]+", "_", s)
        return s.strip("_")
    ep_key = f"{_slug(dom_name)}.{_slug(provider)}"
    print(f"✅ Registered mapper entry point as '{ep_key}'.")

    # Inject per-file canonical stream into streams directory
    streams_path = resolve_streams_dir(proj_path)

    canonical_alias = src_key  # default canonical stream alias
    mapper_ep = ep_key
    # Write a single-file canonical spec into streams directory
    try:
        # Ensure streams_path is a directory path
        streams_dir = streams_path if streams_path.is_dir() else streams_path.parent
        streams_dir.mkdir(parents=True, exist_ok=True)
        cfile = streams_dir / f"{canonical_alias}.yaml"
        data = {
            SOURCE_KEY: src_key,
            MAPPER_KEY: {ENTRYPOINT_KEY: mapper_ep, ARGS_KEY: {}},
        }
        with cfile.open("w", encoding="utf-8") as f:
            yaml.safe_dump(data, f, sort_keys=False, default_flow_style=False)
        print(f"✨ Created canonical spec: {cfile}")
    except Exception as e:
        print(f"❗ Failed to write canonical spec: {e}", file=sys.stderr)



from __future__ import annotations

from datetime import datetime
from typing import Optional
from pydantic import BaseModel, Field, ConfigDict


class ProjectPaths(BaseModel):
    streams: str
    sources: str
    dataset: str


class ProjectGlobals(BaseModel):
    model_config = ConfigDict(extra='allow')

    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None


class ProjectConfig(BaseModel):
    version: int = 1
    paths: ProjectPaths
    globals: ProjectGlobals = Field(default_factory=ProjectGlobals)



from ..constants import LOADERS_GROUP, PARSERS_GROUP
from pathlib import Path
from typing import Optional

from datapipeline.services.scaffold.templates import camel, render
from datapipeline.utils.load import load_yaml

from ..constants import COMPOSED_LOADER_EP
from ..entrypoints import inject_ep
from ..paths import pkg_root, resolve_base_pkg_dir


def _class_prefix(provider: str, dataset: str) -> str:
    """Single place to define class-prefix naming."""
    return f"{camel(provider)}{camel(dataset)}"


def _source_alias(provider: str, dataset: str) -> str:
    return f"{provider}_{dataset}"


def _write_if_missing(path: Path, text: str) -> None:
    """Write file only if it does not exist; echo a friendly message."""
    if not path.exists():
        path.write_text(text)
        print(f"✨ Created: {path}")


def _render_loader_stub(transport: str, loader_class: str,
                        *, fmt: Optional[str]) -> str | None:
    """Render loader stub from Jinja templates for supported transports."""
    if transport == "synthetic":
        return render("loader_synthetic.py.j2", CLASS_NAME=loader_class)
    return None


def _update_ep(toml_text: str, provider: str, dataset: str, pkg_name: str,
               transport: str, parser_class: str, loader_class: str) -> tuple[str, str]:
    """
    Inject parser EP (always). Returns (updated_toml, ep_key).
    """
    ep_key = f"{provider}.{dataset}"
    toml_text = inject_ep(
        toml_text, PARSERS_GROUP, ep_key,
        f"{pkg_name}.sources.{provider}.{dataset}.parser:{parser_class}"
    )
    if transport in {"synthetic"}:
        toml_text = inject_ep(
            toml_text, LOADERS_GROUP, ep_key,
            f"{pkg_name}.sources.{provider}.{dataset}.loader:{loader_class}"
        )
    return toml_text, ep_key


def _loader_ep_and_args(transport: str, fmt: Optional[str], ep_key: str) -> tuple[str, dict]:
    """Return (loader EP name, default args) for the YAML snippet."""
    if transport == "fs":
        args = {
            "transport": "fs",
            "format": fmt or "<FORMAT (csv|json|json-lines)>",
            "path": "<PATH OR GLOB>",
            "glob": False,
            "encoding": "utf-8",
        }
        if fmt == "csv":
            args["delimiter"] = ","
        return COMPOSED_LOADER_EP, args
    if transport == "synthetic":
        return ep_key, {"start": "<ISO8601>", "end": "<ISO8601>", "frequency": "1h"}
    if transport == "url":
        args = {
            "transport": "url",
            "format": fmt or "<FORMAT (json|json-lines|csv)>",
            "url": "<https://api.example.com/data.json>",
            "headers": {},
            "encoding": "utf-8",
        }
        if fmt == "csv":
            args["delimiter"] = ","
        return COMPOSED_LOADER_EP, args
    return ep_key, {}


def create_source(*, provider: str, dataset: str, transport: str,
                  format: Optional[str], root: Optional[Path]) -> None:
    root_dir, name, _ = pkg_root(root)
    base = resolve_base_pkg_dir(root_dir, name)
    src_pkg_dir = base / "sources" / provider / dataset
    src_pkg_dir.mkdir(parents=True, exist_ok=True)
    (src_pkg_dir / "__init__.py").touch(exist_ok=True)

    class_prefix = _class_prefix(provider, dataset)
    dto_class = f"{class_prefix}DTO"
    parser_class = f"{class_prefix}Parser"
    loader_class = f"{class_prefix}DataLoader"

    # DTO
    dto_path = src_pkg_dir / "dto.py"
    _write_if_missing(dto_path, render(
        "dto.py.j2",
        PACKAGE_NAME=name, ORIGIN=provider, DOMAIN=dataset,
        CLASS_NAME=dto_class, time_aware=False
    ))

    # Parser
    parser_path = src_pkg_dir / "parser.py"
    _write_if_missing(parser_path, render(
        "parser.py.j2",
        PACKAGE_NAME=name, ORIGIN=provider, DOMAIN=dataset,
        CLASS_NAME=parser_class, DTO_CLASS=dto_class, time_aware=False
    ))

    # Optional loader stub: synthetic (url uses composed loader by default)
    if transport in {"synthetic"}:
        loader_path = src_pkg_dir / "loader.py"
        stub = _render_loader_stub(transport, loader_class, fmt=format)
        if stub is not None:
            _write_if_missing(loader_path, stub)

    toml_path = root_dir / "pyproject.toml"
    toml_text, ep_key = _update_ep(
        toml_path.read_text(),
        provider, dataset, name,
        transport, parser_class, loader_class
    )
    toml_path.write_text(toml_text)

    alias = _source_alias(provider, dataset)
    loader_ep, loader_args = _loader_ep_and_args(transport, format, ep_key)

    proj_yaml = root_dir / "config" / "project.yaml"
    sources_dir = root_dir / "config" / "sources"
    try:
        if proj_yaml.exists():
            proj = load_yaml(proj_yaml)
            paths = proj.get("paths") if isinstance(proj, dict) else None
            if isinstance(paths, dict):
                sdir = paths.get("sources")
                if isinstance(sdir, str) and sdir:
                    sources_dir = (Path(sdir)
                                   if Path(sdir).is_absolute()
                                   else (root_dir / sdir))
                else:
                    s = paths.get("streams")
                    if isinstance(s, str) and s:
                        sp = Path(s)
                        sources_dir = (sp.parent if sp.is_absolute() else (
                            root_dir / sp).parent) / "sources"
    except Exception:
        # Best-effort resolution; ignore errors and use default
        pass
    sources_dir.mkdir(parents=True, exist_ok=True)
    src_cfg_path = sources_dir / f"{alias}.yaml"
    if not src_cfg_path.exists():
        src_cfg_path.write_text(render(
            "source.yaml.j2",
            parser_ep=ep_key,
            parser_args={},
            loader_ep=loader_ep,
            loader_args=loader_args,
            composed_loader_ep=COMPOSED_LOADER_EP,
        ))
        print(f"✨ Created: {src_cfg_path}")

    # No inline instructions; per-source YAML file has been created above.




from datetime import datetime, timezone
from dataclasses import dataclass, field
from typing import Any


@dataclass
class Record:
    # Canonical payload for pipelines; may be numeric or categorical.
    value: Any


@dataclass
class TimeFeatureRecord(Record):
    time: datetime

    def __post_init__(self):
        if self.time.tzinfo is None:
            raise ValueError("TimeFeatureRecord.time must be timezone-aware")
        self.time = self.time.astimezone(timezone.utc)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Iterator, List, Dict, Optional
from urllib.request import Request, urlopen
from urllib.error import URLError, HTTPError


class TextSource(ABC):
    """Abstract transport that yields text line streams per resource.

    Each item from `streams()` is an iterable over text lines (no newlines).
    """

    @abstractmethod
    def streams(self) -> Iterator[Iterable[str]]:
        pass


class FsFileSource(TextSource):
    def __init__(self, path: str, *, encoding: str = "utf-8"):
        self.path = path
        self.encoding = encoding

    def streams(self) -> Iterator[Iterable[str]]:
        def _iter() -> Iterator[str]:
            with open(self.path, "r", encoding=self.encoding) as f:
                for line in f:
                    yield line
        yield _iter()


class FsGlobSource(TextSource):
    def __init__(self, pattern: str, *, encoding: str = "utf-8"):
        import glob as _glob

        self.pattern = pattern
        self.encoding = encoding
        self._files: List[str] = sorted(_glob.glob(pattern))

    def streams(self) -> Iterator[Iterable[str]]:
        def _iter(path: str) -> Iterator[str]:
            with open(path, "r", encoding=self.encoding) as f:
                for line in f:
                    yield line
        for p in self._files:
            yield _iter(p)


class UrlSource(TextSource):
    def __init__(self, url: str, *, headers: Optional[Dict[str, str]] = None, encoding: str = "utf-8"):
        self.url = url
        self.headers = dict(headers or {})
        self.encoding = encoding

    def streams(self) -> Iterator[Iterable[str]]:
        req = Request(self.url, headers=self.headers)
        try:
            with urlopen(req) as resp:
                data = resp.read()
        except (URLError, HTTPError) as e:
            raise RuntimeError(f"failed to fetch {self.url}: {e}") from e

        text = data.decode(self.encoding, errors="strict")
        lines = text.splitlines()
        yield iter(lines)



from datapipeline.domain.record import Record
from dataclasses import dataclass
from typing import Union


@dataclass
class FeatureRecord:
    record: Union[Record, list[Record]]
    feature_id: str
    group_key: tuple



from __future__ import annotations

from pathlib import Path

from datapipeline.utils.load import load_yaml
from datapipeline.config.project import ProjectConfig


def read_project(project_yaml: Path) -> ProjectConfig:
    data = load_yaml(project_yaml)
    return ProjectConfig.model_validate(data)


def _project_root(project_yaml: Path) -> Path:
    return project_yaml.parent.parent


def streams_dir(project_yaml: Path) -> Path:
    cfg = read_project(project_yaml)
    p = Path(cfg.paths.streams)
    if not p.is_absolute():
        p = _project_root(project_yaml) / p
    if not p.exists() or not p.is_dir():
        raise FileNotFoundError(f"streams dir not found: {p}")
    return p


def sources_dir(project_yaml: Path) -> Path:
    cfg = read_project(project_yaml)
    p = Path(cfg.paths.sources)
    if not p.is_absolute():
        p = _project_root(project_yaml) / p
    if not p.exists() or not p.is_dir():
        raise FileNotFoundError(f"sources dir not found: {p}")
    return p



from datapipeline.utils.load import load_ep
from datapipeline.plugins import PARSERS_EP, LOADERS_EP, MAPPERS_EP
from datapipeline.sources.models.source import Source
from datapipeline.config.catalog import RawSourceSpec, EPArgs


def build_source_from_spec(spec: RawSourceSpec) -> Source:
    P = load_ep(PARSERS_EP, spec.parser.entrypoint)
    L = load_ep(LOADERS_EP, spec.loader.entrypoint)
    return Source(loader=L(**spec.loader.args), parser=P(**spec.parser.args))


def build_mapper_from_spec(spec: EPArgs | None):
    if not spec or not spec.entrypoint:
        from datapipeline.mappers.noop import identity
        return identity, {}
    fn = load_ep(MAPPERS_EP, spec.entrypoint)
    return fn, dict(spec.args or {})



from typing import Union, List, Any


class RecordKeyGenerator:
    """
    Generates unique feature keys by appending suffixes from expand_by fields.
    """

    def __init__(self, partition_by: Union[str, List[str], None]):
        self.partition_by = partition_by

    def generate(self, base_id: str, record: Any) -> str:
        if not self.partition_by:
            return base_id
        if isinstance(self.partition_by, str):
            suffix = getattr(record, self.partition_by)
        else:
            suffix = "__".join(str(getattr(record, f))
                               for f in self.partition_by)
        return f"{base_id}__{suffix}"



from datapipeline.services.paths import pkg_root, resolve_base_pkg_dir
from datapipeline.services.project_paths import sources_dir as sources_dir_from_project


def handle(subcmd: str) -> None:
    root_dir, name, pyproject = pkg_root(None)
    if subcmd == "sources":
        # Discover sources by scanning sources_dir for YAML files
        proj_path = root_dir / "config" / "project.yaml"
        sources_dir = sources_dir_from_project(proj_path)
        if sources_dir.exists():
            aliases = sorted(p.stem for p in sources_dir.glob("*.y*ml"))
            for a in aliases:
                print(a)
    elif subcmd == "domains":
        base = resolve_base_pkg_dir(root_dir, name)
        dom_dir = base / "domains"
        if dom_dir.exists():
            names = sorted(p.name for p in dom_dir.iterdir()
                           if p.is_dir() and (p / "model.py").exists())
            for k in names:
                print(k)



"""Helpers for assembling the different pipeline stages."""

from __future__ import annotations

import heapq
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import Any, Tuple

from datapipeline.config.dataset.feature import BaseRecordConfig, FeatureRecordConfig
from datapipeline.config.dataset.group_by import GroupBy
from datapipeline.domain.feature import FeatureRecord
from datapipeline.domain.vector import Vector
from datapipeline.pipeline.stages import feature_stage, record_stage, vector_stage


def build_record_pipeline(
    cfg: BaseRecordConfig,
    open_stream: Callable[[str], Iterable[Any]],
) -> Iterator[Any]:
    """Open a configured stream and apply record-level filters/transforms."""

    raw = open_stream(cfg.stream)
    return record_stage(raw, cfg.filters, cfg.transforms)


def build_feature_pipeline(
    cfg: FeatureRecordConfig,
    group_by: GroupBy,
    open_stream: Callable[[str], Iterable[Any]],
) -> Iterator[FeatureRecord]:
    """Build the feature-level stream for a single feature configuration."""

    rec = build_record_pipeline(cfg, open_stream)
    return feature_stage(rec, cfg, group_by)


def build_vector_pipeline(
    configs: Sequence[FeatureRecordConfig],
    group_by: GroupBy,
    open_stream: Callable[[str], Iterable[Any]],
) -> Iterator[Tuple[Any, Vector]]:
    """Merge feature streams and yield grouped vectors ready for export."""

    streams = [build_feature_pipeline(c, group_by, open_stream) for c in configs]
    merged = heapq.merge(*streams, key=lambda fr: fr.group_key)
    return vector_stage(merged)



from typing import Iterator
from datapipeline.domain.feature import FeatureRecord
from itertools import groupby
from collections import deque


class TimeWindowTransformer:
    def __init__(self, size: int, stride: int = 1):
        self.size = size
        self.stride = stride

    def apply(self, stream: Iterator[FeatureRecord]) -> Iterator[FeatureRecord]:
        """Assumes input is pre-sorted by (feature_id, record.time).

        Produces sliding windows per feature_id. Each output FeatureRecord has
        a list[Record] in `record` and carries the current group_key.
        """
        grouped = groupby(stream, key=lambda fr: fr.feature_id)

        for feature_id, records in grouped:
            window = deque(maxlen=self.size)
            step = 0
            for fr in records:
                window.append(fr)
                if len(window) == self.size and step % self.stride == 0:
                    yield FeatureRecord(
                        record=[r.record for r in window],
                        feature_id=feature_id,
                        group_key=fr.group_key,
                    )
                step += 1



from __future__ import annotations

from typing import Any, Dict

from datapipeline.sources.composed_loader import ComposedRawLoader
from datapipeline.sources.transports import FsFileSource, FsGlobSource, UrlSource
from datapipeline.sources.decoders import CsvDecoder, JsonDecoder, JsonLinesDecoder


def build_loader(*, transport: str, format: str | None = None, **kwargs: Any) -> ComposedRawLoader:
    """Factory entrypoint that composes a transport and a decoder.

    Args (by transport/format):
      transport: "fs" | "url"
      format: "csv" | "json" | "json-lines" (required for fs/url)
      fs: path (str), glob (bool, optional), encoding (str, default utf-8), delimiter (csv only)
      url: url (str), headers (dict, optional), encoding (str, default utf-8)
    """

    t = (transport or "").lower()
    fmt = (format or "").lower()

    # Build source
    if t == "fs":
        path = kwargs.get("path")
        if not path:
            raise ValueError("fs transport requires 'path'")
        encoding = kwargs.get("encoding", "utf-8")
        use_glob = bool(kwargs.get("glob", False))
        source = FsGlobSource(path, encoding=encoding) if use_glob else FsFileSource(path, encoding=encoding)
    elif t == "url":
        url = kwargs.get("url")
        if not url:
            raise ValueError("url transport requires 'url'")
        headers: Dict[str, str] = dict(kwargs.get("headers") or {})
        encoding = kwargs.get("encoding", "utf-8")
        source = UrlSource(url, headers=headers, encoding=encoding)
    else:
        raise ValueError(f"unsupported transport: {transport}")

    # Build decoder
    if fmt == "csv":
        delimiter = kwargs.get("delimiter", ";")
        decoder = CsvDecoder(delimiter=delimiter)
    elif fmt == "json":
        decoder = JsonDecoder()
    elif fmt == "json-lines":
        decoder = JsonLinesDecoder()
    else:
        raise ValueError(f"unsupported format for composed loader: {format}")

    allow_net = bool(kwargs.get("count_by_fetch", False))
    return ComposedRawLoader(source, decoder, allow_network_count=allow_net)



from pathlib import Path
from typing import Optional

from datapipeline.services.bootstrap import bootstrap
from datapipeline.config.dataset.loader import load_dataset
from datapipeline.pipeline.pipelines import build_vector_pipeline
from datapipeline.cli.openers import open_canonical_stream_visual
from datapipeline.analysis.vector_analyzer import VectorStatsCollector


def analyze(project: str, limit: Optional[int] = None) -> None:
    project_path = Path(project)
    dataset = load_dataset(project_path, "vectors")
    bootstrap(project_path)

    expected_feature_ids = [cfg.feature_id for cfg in (dataset.features or [])]
    if not expected_feature_ids:
        print("(no features configured; nothing to analyze)")
        return

    collector = VectorStatsCollector(expected_feature_ids)

    count = 0
    for group_key, vector in build_vector_pipeline(
        dataset.features, dataset.group_by, open_canonical_stream_visual
    ):
        collector.update(group_key, vector.values)
        count += 1
        if limit and count >= limit:
            break

    collector.print_report()



from typing import Iterator, Any
from datapipeline.streams.canonical import canonical_entry
from datapipeline.streams.raw import open_raw_stream
from datapipeline.cli.visuals import wrap_with_tqdm


def open_canonical_stream_visual(alias: str, show: bool = True) -> Iterator[Any]:
    entry = canonical_entry(alias)
    raw = open_raw_stream(entry.source_alias)
    wrapped = wrap_with_tqdm(raw, stream_alias=alias, show=show)
    return entry.mapper(wrapped, **entry.mapper_args)



from dataclasses import dataclass


@dataclass(frozen=True)
class Point:
    lon: float
    lat: float

    def to_tuple(self) -> tuple[float, float]:
        return (self.lon, self.lat)

    def to_geojson(self) -> dict:
        return {"type": "Point", "coordinates": [self.lon, self.lat]}



import argparse

from datapipeline.cli.commands.run import handle_prep, handle_serve
from datapipeline.cli.commands.analyze import analyze as handle_analyze
from datapipeline.cli.commands.plugin import station as handle_station
from datapipeline.cli.commands.source import handle as handle_source
from datapipeline.cli.commands.domain import handle as handle_domain
from datapipeline.cli.commands.link import handle as handle_link
from datapipeline.cli.commands.list_ import handle as handle_list
from datapipeline.cli.commands.filter import handle as handle_filter


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="jerry",
        description="Mixology-themed CLI for building and serving data pipelines.",
    )
    sub = parser.add_subparsers(dest="cmd", required=True)

    # prep (debug mode with visuals)
    p_prep = sub.add_parser(
        "prep",
        help="run pipeline stages with visual progress",
    )
    prep_sub = p_prep.add_subparsers(dest="prep_cmd", required=True)
    prep_steps = {
        "pour": "preview record-stage output",
        "build": "inspect feature-stage output",
        "stir": "examine vector-stage output",
    }
    for step, help_text in prep_steps.items():
        sp = prep_sub.add_parser(step, help=help_text)
        sp.add_argument(
            "--project", "-p", default="config/project.yaml", help="path to project.yaml"
        )
        sp.add_argument("--limit", "-n", type=int, default=20)

    # serve (production run, no visuals)
    p_serve = sub.add_parser(
        "serve",
        help="produce vectors without progress visuals",
    )
    p_serve.add_argument(
        "--project", "-p", default="config/project.yaml", help="path to project.yaml"
    )
    p_serve.add_argument(
        "--limit", "-n", type=int, default=None,
        help="optional cap on the number of vectors to emit",
    )
    p_serve.add_argument(
        "--output", "-o", default="print",
        help="output destination: 'print', 'stream', or a file ending in .pt",
    )

    # taste (analysis)
    p_taste = sub.add_parser(
        "taste",
        help="analyze vector completeness and feature stats",
    )
    p_taste.add_argument(
        "--project", "-p", default="config/project.yaml", help="path to project.yaml"
    )
    p_taste.add_argument("--limit", "-n", type=int, default=None)

    # distillery (sources)
    p_dist = sub.add_parser(
        "distillery",
        help="add or list raw sources",
    )
    dist_sub = p_dist.add_subparsers(dest="dist_cmd", required=True)
    p_dist_add = dist_sub.add_parser(
        "add",
        help="create a provider+dataset source",
        description=(
            "Scaffold a source using transport + format.\n\n"
            "Examples:\n"
            "  fs CSV:        -t fs  -f csv\n"
            "  fs NDJSON:     -t fs  -f json-lines\n"
            "  URL JSON:      -t url -f json\n"
            "  Synthetic:     -t synthetic\n\n"
            "Note: set 'glob: true' in the generated YAML if your 'path' contains wildcards."
        ),
    )
    p_dist_add.add_argument("--provider", "-p", required=True)
    p_dist_add.add_argument("--dataset", "-d", required=True)
    p_dist_add.add_argument(
        "--transport", "-t",
        choices=["fs", "url", "synthetic"],
        required=True,
        help="how data is accessed: fs/url/synthetic",
    )
    p_dist_add.add_argument(
        "--format", "-f",
        choices=["csv", "json", "json-lines"],
        help="data format for fs/url transports (ignored otherwise)",
    )
    dist_sub.add_parser("list", help="list known sources")

    # spirit (domains)
    p_spirit = sub.add_parser(
        "spirit",
        help="add or list domains",
    )
    spirit_sub = p_spirit.add_subparsers(dest="spirit_cmd", required=True)
    p_spirit_add = spirit_sub.add_parser(
        "add",
        help="create a domain",
        description=(
            "Create a domain package. Defaults to Record base. "
            "Use --time-aware to base on TimeFeatureRecord (adds 'time' and 'value' fields)."
        ),
    )
    p_spirit_add.add_argument("--domain", "-d", required=True)
    p_spirit_add.add_argument(
        "--time-aware",
        "-t",
        action="store_true",
        help="use TimeFeatureRecord base (UTC-aware 'time' + 'value' fields) instead of Record",
    )
    spirit_sub.add_parser("list", help="list known domains")

    # contract (link source ↔ domain)
    p_contract = sub.add_parser(
        "contract",
        help="link a distillery source to a spirit domain",
    )
    p_contract.add_argument("--time-aware", "-t", action="store_true")

    # station (plugin scaffolding)
    p_station = sub.add_parser(
        "station",
        help="scaffold plugin workspaces",
    )
    station_sub = p_station.add_subparsers(dest="station_cmd", required=True)
    p_station_init = station_sub.add_parser(
        "init", help="create a plugin skeleton")
    p_station_init.add_argument("--name", "-n", required=True)
    p_station_init.add_argument("--out", "-o", default=".")

    # filter (unchanged helper)
    p_filt = sub.add_parser("filter", help="manage filters")
    filt_sub = p_filt.add_subparsers(dest="filter_cmd", required=True)
    p_filt_create = filt_sub.add_parser(
        "create", help="create a filter function")
    p_filt_create.add_argument(
        "--name", "-n", required=True,
        help="filter entrypoint name and function/module name",
    )

    args = parser.parse_args()

    if args.cmd == "prep":
        handle_prep(action=args.prep_cmd,
                    project=args.project, limit=args.limit)
        return

    if args.cmd == "serve":
        handle_serve(
            project=args.project,
            limit=getattr(args, "limit", None),
            output=args.output,
        )
        return

    if args.cmd == "taste":
        handle_analyze(project=args.project,
                       limit=getattr(args, "limit", None))
        return

    if args.cmd == "distillery":
        if args.dist_cmd == "list":
            handle_list(subcmd="sources")
        else:
            handle_source(
                subcmd="add",
                provider=getattr(args, "provider", None),
                dataset=getattr(args, "dataset", None),
                transport=getattr(args, "transport", None),
                format=getattr(args, "format", None),
            )
        return

    if args.cmd == "spirit":
        if args.spirit_cmd == "list":
            handle_list(subcmd="domains")
        else:
            handle_domain(
                subcmd="add",
                domain=getattr(args, "domain", None),
                time_aware=getattr(args, "time_aware", False),
            )
        return

    if args.cmd == "contract":
        handle_link(time_aware=getattr(args, "time_aware", False))
        return

    if args.cmd == "station":
        handle_station(
            subcmd=args.station_cmd,
            name=getattr(args, "name", None),
            out=getattr(args, "out", "."),
        )
        return

    if args.cmd == "filter":
        handle_filter(subcmd=args.filter_cmd, name=getattr(args, "name", None))
        return



"""Dataset feature configuration models."""

from __future__ import annotations

from typing import Any, List, Optional, Union

from pydantic import BaseModel


class BaseRecordConfig(BaseModel):
    """Configuration shared by record- and feature-level pipelines."""

    stream: str
    feature_id: str
    partition_by: Optional[Union[str, List[str]]] = None
    filters: Optional[List[Any]] = None
    transforms: Optional[List[Any]] = None


class FeatureRecordConfig(BaseRecordConfig):
    """Configuration for feature-level pipelines (post record stage)."""

    feature_transforms: Optional[List[Any]] = None
    sequence_transforms: Optional[List[Any]] = None



from datetime import datetime


def floor_time_to_resolution(ts: datetime, resolution: str) -> datetime:
    if resolution == "1h":
        return ts.replace(minute=0, second=0, microsecond=0)
    elif resolution == "10min":
        floored_minute = (ts.minute // 10) * 10
        return ts.replace(minute=floored_minute, second=0, microsecond=0)
    raise ValueError(f"Unsupported granularity: {resolution}")



from typing import Dict, Optional, Any
from pydantic import BaseModel, Field


class EPArgs(BaseModel):
    entrypoint: str
    args: Dict[str, Any] = Field(default_factory=dict)


class RawSourceSpec(BaseModel):
    parser: EPArgs
    loader: EPArgs


class CanonicalSpec(BaseModel):
    source: str
    mapper: Optional[EPArgs] = None


class StreamsConfig(BaseModel):
    raw: Dict[str, RawSourceSpec] = Field(default_factory=dict)
    canonical: Dict[str, CanonicalSpec] = Field(default_factory=dict)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable, Iterator, Dict, Any, Optional
import csv
import json
from io import StringIO


class Decoder(ABC):
    @abstractmethod
    def decode(self, lines: Iterable[str]) -> Iterator[Dict[str, Any]]:
        pass

    def count(self, lines: Iterable[str]) -> Optional[int]:
        """Optional fast count of rows for the given stream.

        Default returns None. Subclasses may override for better visuals.
        Note: This will consume the provided iterable.
        """
        return None


class CsvDecoder(Decoder):
    def __init__(self, *, delimiter: str = ";"):
        self.delimiter = delimiter

    def decode(self, lines: Iterable[str]) -> Iterator[Dict[str, Any]]:
        # Stream directly from the line iterator; no buffering
        reader = csv.DictReader(lines, delimiter=self.delimiter)
        for row in reader:
            yield row

    def count(self, lines: Iterable[str]) -> Optional[int]:
        return sum(1 for _ in csv.DictReader(lines, delimiter=self.delimiter))


class JsonDecoder(Decoder):
    def decode(self, lines: Iterable[str]) -> Iterator[Dict[str, Any]]:
        text = "\n".join(lines)
        data = json.loads(text)
        if isinstance(data, list):
            for item in data:
                yield item
        else:
            # Yield a single object as one row
            yield data

    def count(self, lines: Iterable[str]) -> Optional[int]:
        text = "\n".join(lines)
        data = json.loads(text)
        return len(data) if isinstance(data, list) else 1


class JsonLinesDecoder(Decoder):
    def decode(self, lines: Iterable[str]) -> Iterator[Dict[str, Any]]:
        for line in lines:
            s = line.strip()
            if not s:
                continue
            yield json.loads(s)

    def count(self, lines: Iterable[str]) -> Optional[int]:
        return sum(1 for s in lines if s.strip())



from datapipeline.services.scaffold.source import create_source


def handle(subcmd: str, provider: str | None, dataset: str | None,
           transport: str | None = None, format: str | None = None) -> None:
    if subcmd in {"create", "add"}:
        if not provider or not dataset:
            print("❗ --provider and --dataset are required")
            raise SystemExit(2)
        if not transport:
            print("❗ --transport is required (fs|url|synthetic)")
            raise SystemExit(2)
        if transport in {"fs", "url"} and not format:
            print("❗ --format is required for fs/url transports (csv|json|json-lines)")
            raise SystemExit(2)
        create_source(provider=provider, dataset=dataset,
                      transport=transport, format=format, root=None)



from typing import Dict, Iterator
from datapipeline.sources.models.source import Source

_SOURCES: Dict[str, Source] = {}


def register_source(alias: str, source: Source) -> None:
    _SOURCES[alias] = source


def get_source(alias: str) -> Source:
    return _SOURCES[alias]


def open_raw_stream(alias: str) -> Iterator[object]:
    return get_source(alias).stream()



from typing import Iterator, Optional
from tqdm import tqdm
from typing import Any

from pathlib import Path
from urllib.parse import urlparse
from datapipeline.streams.canonical import canonical_entry
from datapipeline.streams.raw import get_source
from datapipeline.sources.models.loader import FileLoader, SyntheticLoader, UrlLoader
from datapipeline.sources.models.loader import RawDataLoader
from datapipeline.sources.composed_loader import ComposedRawLoader
from datapipeline.sources.transports import FsFileSource, FsGlobSource, UrlSource
from datapipeline.sources.decoders import CsvDecoder, JsonDecoder, JsonLinesDecoder


def unit_for_loader(loader) -> str:
    if isinstance(loader, SyntheticLoader):
        return "tick"
    if isinstance(loader, ComposedRawLoader):
        dec = getattr(loader, "decoder", None)
        if isinstance(dec, CsvDecoder):
            return "row"
        if isinstance(dec, (JsonDecoder, JsonLinesDecoder)):
            return "item"
    if isinstance(loader, FileLoader):
        return "item"
    if isinstance(loader, UrlLoader):
        return "item"
    return "record"


def build_source_label(loader: RawDataLoader) -> str:
    if isinstance(loader, SyntheticLoader):
        try:
            gen_name = loader.generator.__class__.__name__
        except Exception:
            gen_name = loader.__class__.__name__
        return "Generating data with " + gen_name
    if isinstance(loader, ComposedRawLoader):
        src = getattr(loader, "source", None)
        if isinstance(src, (FsFileSource, FsGlobSource)):
            name = str(getattr(src, "pattern", getattr(src, "path", "")))
            label = Path(name).name if (
                name and "*" not in name) else (name or "fs")
            return f"Loading data from: {label}"
        if isinstance(src, UrlSource):
            host = urlparse(src.url).netloc or "http"
            return f"Downloading data from: @{host}"
    return loader.__class__.__name__


def icon_for_loader(loader) -> str:
    if isinstance(loader, SyntheticLoader):
        gen = getattr(loader, "generator", None)
        try:
            if gen is not None:
                if any(hasattr(gen, attr) for attr in ("start", "end", "frequency")):
                    return "🕒"
                if hasattr(gen, "seed"):
                    return "🎲"
        except Exception:
            pass
        return "✨"
    if isinstance(loader, ComposedRawLoader):
        src = getattr(loader, "source", None)
        if isinstance(src, (FsFileSource, FsGlobSource)):
            return "📄"
        if isinstance(src, UrlSource):
            return "🌐"
        return "📦"


def _progress_bits_for_feature(stream_alias) -> tuple[str, Optional[int], str]:
    raw_alias = canonical_entry(stream_alias).source_alias
    src = get_source(raw_alias)
    loader = src.loader

    desc = f"{icon_for_loader(loader)} {build_source_label(loader)}"
    unit = unit_for_loader(loader)
    try:
        total = src.count() if hasattr(src, "count") else None
    except Exception:
        total = None
    return desc, total, unit


def wrap_with_tqdm(it: Iterator[Any], stream_alias, show: bool = True) -> Iterator[Any]:
    if not show:
        return it
    desc, total, unit = _progress_bits_for_feature(stream_alias)
    return tqdm(it, total=total, desc=desc, unit=unit, dynamic_ncols=True, mininterval=0.0, miniters=1, leave=True,)
